#
#  Aura Engine
#
#  Playout Daemon for autoradio project
#
#
#  Copyright (C) 2017-2018 Gottfried Gaisbauer <gottfried.gaisbauer@servus.at>
#
#  This file is part of engine.
#
#  engine is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  any later version.
#
#  engine is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with engine. If not, see <http://www.gnu.org/licenses/>.
#


import time
import json
import datetime
import decimal
import traceback
import sqlalchemy
import logging
import threading

from operator import attrgetter

from modules.database.model import AuraDatabaseModel, Schedule, Playlist, PlaylistEntry, PlaylistEntryMetaData, SingleEntry, SingleEntryMetaData, TrackService

from modules.base.exceptions import NoActiveScheduleException, NoActiveEntryException, LoadSourceException
from modules.base.enum import Channel, ChannelType, TimerType, TransitionType, EntryQueueState, EntryPlayState
from modules.base.utils import SimpleUtil, TerminalColors, EngineUtil
from modules.communication.redis.messenger import RedisMessenger
from modules.scheduling.calendar import AuraCalendarService
from modules.scheduling.fallback_manager import FallbackManager


def alchemyencoder(obj):
    """JSON encoder function for SQLAlchemy special classes."""
    if isinstance(obj, datetime.date):
        return obj.isoformat()
    elif isinstance(obj, decimal.Decimal):
        return float(obj)
    elif isinstance(obj, sqlalchemy.orm.state.InstanceState):
        return ""
    elif isinstance(obj, Schedule):
        return json.dumps([obj._asdict()], default=alchemyencoder)
    else:
        return str(obj)


# ------------------------------------------------------------------------------------------ #
class AuraScheduler(threading.Thread):
    """
    Aura Scheduler Class

    - Retrieves data from Steering and Tank
    - Stores and fires events for LiquidSoap

    Attributes:
        config (AuraConfig):                    Holds the Engine Configuration
        logger:                                 The logger
        exit_event(threading.Event):            Used to exit the thread if requested                             
        soundsystem:                            Manages the audio streams via LiquidSoap 
        last_successful_fetch (datetime):       Stores the last time a fetch from Steering/Tank was successful

        programme:                              The current radio programme to be played as defined in the local engine database
        active_entry(Show, Track):              This is a Tuple consisting of the currently played `Show` and `Track`
        message_timer(List<threading.Timer>):   The timer queue of Liquidsoap commands for playlists/entries to be played
    """
    redismessenger = None
    job_result = {}

    config = None
    logger = None
    exit_event = None
    soundsystem = None
    last_successful_fetch = None
    programme = None
    message_timer = []
    fallback_manager = None
    client = None
    is_initialized = None


    def __init__(self, config, soundsystem, func_on_init):
        """
        Constructor

        Args:
            config (AuraConfig):        Reads the engine configuration
            soundsystem (SoundSystem):  The soundsystem to play the schedule on
            func_on_init (Function):    The function to be called when the scheduler is initialized
        """
        self.config = config
        self.logger = logging.getLogger("AuraEngine")

        # self.init_error_messages()
        self.init_database()
        self.fallback_manager = FallbackManager(config, self.logger, self)
        self.redismessenger = RedisMessenger(config)
        self.soundsystem = soundsystem
        self.soundsystem.scheduler = self

        # Scheduler Initialization
        self.is_initialized = False
        self.func_on_initialized = func_on_init

        # init threading
        threading.Thread.__init__(self)

        # init messenger.. FIXME probably not needed anymore
        self.redismessenger.set_channel('scheduler')
        self.redismessenger.set_section('execjob')
        #self.redismessenger.send('Scheduler started', '0000', 'success', 'initApp', None, 'appinternal')

        # Create exit event
        self.exit_event = threading.Event()
        self.start()



    def run(self):
        """
        Called when thread is started via `start()`. It does following:
        
            1. `self.fetch_new_programme()` periodically from the API depending on the `fetching_frequency` defined in the engine configuration. 
            2. Loads the latest programme from the database and sets the instance state `self.programme` with current schedules.
            3. Queues all playlists of the programm, if the soundssystem is ready to accept commands.
        """
        while not self.exit_event.is_set():
            try:
                seconds_to_wait = int(self.config.get("fetching_frequency"))
                self.logger.info(SimpleUtil.cyan("== start fetching new schedules =="))
                next_time = datetime.datetime.now() + datetime.timedelta(seconds=seconds_to_wait)
                self.logger.info("Fetch new programmes every %ss. Next fetching in %ss." % (str(seconds_to_wait), str(next_time)))
                self.fetch_new_programme()

                # The scheduler is ready
                if not self.is_initialized:
                    self.is_initialized = True
                    if self.func_on_initialized:
                        self.func_on_initialized()

                # The soundsystem is ready
                if self.soundsystem.is_ready():
                    self.queue_programme()
            except Exception as e:
                self.logger.critical(SimpleUtil.red("Unhandled error while fetching & scheduling new programme! (%s)" % str(e)), e)

            self.clean_timer_queue()
            self.print_timer_queue()
            self.exit_event.wait(seconds_to_wait)



#
#   PUBLIC METHODS
#


    def on_ready(self):
        """
        Called when the soundsystem is ready.
        """
        self.queue_programme()
        self.logger.info(self.get_ascii_programme())
        self.play_active_entry()


    def play_active_entry(self):
        """
        Plays the entry scheduled for the very current moment and forwards to the scheduled position in time. 
        Usually called when the Engine boots.
        """
        sleep_offset = 10

        active_entry = self.get_active_entry()
        if not active_entry:
            raise NoActiveScheduleException

        # In case of a file-system source, we need to fast-foward to the current marker as per schedule
        if active_entry.type == ChannelType.FILESYSTEM:

            # Calculate the seconds we have to fast-forward
            now_unix = self.get_virtual_now()
            seconds_to_seek = now_unix - active_entry.start_unix

            # If the seek exceeds the length of the current track, 
            # there's no need to do anything - the scheduler takes care of the rest
            if (seconds_to_seek + sleep_offset) > active_entry.duration:
                self.logger.info("The FFWD [>>] range exceeds the length of the entry. Drink some tea and wait for the sound of the next entry.")
            else:
                # Pre-roll and play active entry
                self.soundsystem.preroll(active_entry)
                self.soundsystem.play(active_entry, TransitionType.FADE)

                # Check if this is the last item of the schedule
                if active_entry.end_unix > active_entry.playlist.schedule.end_unix:
                    self.queue_end_of_schedule(active_entry, True)

                # Fast-forward to the scheduled position
                if seconds_to_seek > 0:
                    # Without plenty of timeout (10s) the seek doesn't work
                    seconds_to_seek += sleep_offset
                    time.sleep(sleep_offset)
                    self.logger.info("Going to fast-forward %s seconds" % seconds_to_seek)
                    self.soundsystem.enable_transaction()
                    response = self.soundsystem.playlist_seek(active_entry.channel, seconds_to_seek)
                    self.soundsystem.disable_transaction()
                    self.logger.info("LiquidSoap seek response: " + response)
        
        elif active_entry.type == ChannelType.HTTP \
            or active_entry.type == ChannelType.HTTPS \
            or active_entry.type == ChannelType.LIVE:

                # Pre-roll and play active entry
                self.soundsystem.preroll(active_entry)
                self.soundsystem.play(active_entry, TransitionType.FADE)

                self.queue_end_of_schedule(active_entry, True)

        else:
            self.logger.critical("Unknown Entry Type: %s" % active_entry)




    def get_active_entry(self):
        """
        Retrieves the current `PlaylistEntry` which should be played as per programme. 

        Returns:
            (PlaylistEntry): The track which is (or should) currently being played
        """
        now_unix = self.get_virtual_now()

        # Load programme if necessary
        if not self.programme:
            self.load_programme_from_db()

        # Check for scheduled playlist
        current_schedule, current_playlist = self.get_active_playlist()

        if not current_schedule:
            self.logger.warning("There's no active schedule")
            return None
        if not current_playlist:
            self.logger.warning("There's no active playlist for a current schedule. Most likely the playlist finished before the end of the schedule.")
            return None

        # Iterate over playlist entries and store the current one
        current_entry = None
        for entry in current_playlist.entries:
            self.logger.info(entry)
            if entry.start_unix <= now_unix and now_unix <= entry.end_unix:
                current_entry = entry
                break       
      
        if not current_entry:
            # Nothing playing ... fallback will kick-in
            self.logger.warning("There's no entry scheduled for playlist '%s' at time %s" % (str(current_playlist), SimpleUtil.fmt_time(now_unix)))
            return None

        return current_entry



    def get_active_playlist(self):
        """
        Retrieves the schedule and playlist currently to be played as per 
        schedule. If the current point in time has no playlist assigned,
        only the matching schedule is returned.

        Returns:
            (Schedule, Playlist): The current schedule and playlist tuple.
        """
        current_schedule = None
        current_playlist = None
        now_unix = self.get_virtual_now()

        # Iterate over all shows and playlists and find the one to be played right now
        if self.programme:
            for schedule in self.programme:
                if schedule.start_unix <= now_unix and now_unix < schedule.end_unix:
                    current_schedule = schedule
                    for playlist in schedule.playlist:
                        if playlist.start_unix <= now_unix and now_unix < playlist.end_unix:
                            current_playlist = playlist 
                            break
                    break

        return current_schedule, current_playlist



    # FIXME Review relevance.
    def get_act_programme_as_string(self):
        """
        Fetches the latest programme and returns it as `String`.
        Also used by `ServerRedisAdapter`.

        Return:
            (String):       Programme

        Raises:
            (Exception):    In case the programme cannot be converted to String
        """
        programme_as_string = ""

        if self.programme is None or len(self.programme) == 0:
            self.fetch_new_programme()

        try:
            programme_as_string = json.dumps([p._asdict() for p in self.programme], default=alchemyencoder)
            # FIXME Change to more specific exception
        except Exception as e:
            self.logger.error("Cannot transform programme into JSON String. Reason: " + str(e))
            traceback.print_exc()

        return programme_as_string



    def print_timer_queue(self):
        """
        Prints the current timer queue i.e. playlists in the queue to be played.
        """
        message_queue = ""
        messages = sorted(self.message_timer, key=attrgetter('diff'))
        if not messages:
            self.logger.warning("There's nothing in the Timer Queue!")
        else:
            for msg in messages:
                message_queue += str(msg)+"\n"

            self.logger.info("Timer queue: \n" + message_queue) 



    def clean_timer_queue(self):
        """
        Removes inactive timers from the queue.
        """
        len_before = len(self.message_timer)
        self.message_timer[:] = [m for m in self.message_timer if m.is_alive()]
        len_after = len(self.message_timer)
        self.logger.info("Removed %s finished timer objects from queue" % (len_before - len_after))



    # ------------------------------------------------------------------------------------------ #
    def set_next_file_for(self, playlistname):
        self.logger.critical("HAVE TO <SET> NEXT FILE FOR: " + playlistname)
        self.logger.critical(str(self.get_active_entry(0)))

        if playlistname == "station":
            file = "/home/david/Code/aura/engine2/testing/content/ernie_mayne_sugar.mp3"
        elif playlistname == "timeslot":
            file = "/home/david/Code/aura/engine2/testing/content/ernie_mayne_sugar.mp3"
        elif playlistname == "show":
            file = "/home/david/Code/aura/engine2/testing/content/ernie_mayne_sugar.mp3"
        else:
            file = ""
            self.logger.critical("Should set next fallback file for " + playlistname + ", but this playlist is unknown!")

        self.logger.info("Set next fallback file for " + playlistname + ": " + file)
        self.redismessenger.set_next_file_for(playlistname, file)
        return file


    def get_next_file_for(self, fallbackname):
        """
        Evaluates the next **fallback files/playlists** to be played for a given fallback-type.
        Valid fallback-types are:

            * timeslot
            * show
            * station

        Args:
            fallbackname (String):  The name of the fallback-type

        Returns:
            (String):   Absolute path to the file to be played as a fallback.
        """

        file = self.fallback_manager.get_fallback_for(fallbackname)

        if file:
            self.logger.info("Got next file '%s' (fallback type: %s)" % (file, fallbackname))

        #set_next_file_thread = SetNextFile(fallbackname, show)
        #set_next_file_thread.start()
        #self.redismessenger.set_next_file_for(playlistname, file)
        return file




    def get_ascii_programme(self):
        """
        Creates a printable version of the current programme (playlists and entries as per schedule) 

        Returns:
            (String):   A ASCII representation of the programme
        """
        active_schedule, active_playlist = self.get_active_playlist()
        playlists = self.get_next_playlists()

        s = "\n\n   PLAYING NOW:"
        s += "\n┌──────────────────────────────────────────────────────────────────────────────────────────────────────"
        if active_schedule:
            s += "\n│   Playing schedule %s         " % active_schedule
            if active_playlist:
                s += "\n│       └── Playlist %s         " % active_playlist
                
                # active_entry = active_playlist.current_entry
                active_entry = self.get_active_entry()

                # Finished entries
                for entry in active_playlist.entries:
                    if active_entry == entry:
                        break
                    else:
                        s += self.build_entry_string("\n│         └── ", entry, True)

                # Entry currently being played
                if active_entry:
                    s += "\n│         └── Entry %s | %s         " % \
                        (str(entry.entry_num+1), SimpleUtil.green("PLAYING > "+str(active_entry)))

                    # Open entries for current playlist
                    rest_of_playlist = active_entry.get_next_entries(False)
                    entries = self.preprocess_entries(rest_of_playlist, False)
                    s += self.build_playlist_string(entries)
                    
            else:
                s += "\n│       └── %s" % (SimpleUtil.red("No playlist active. Did it finish before the end of the schedule?"))
        else:
            s += "\n│   Nothing.           "
        s += "\n└──────────────────────────────────────────────────────────────────────────────────────────────────────"

        s += "\n   PLAYING NEXT:"
        s += "\n┌──────────────────────────────────────────────────────────────────────────────────────────────────────"

        if not playlists:
            s += "\n│   Nothing.         "
        else:
            for next_playlist in playlists:
                s += "\n│   Queued schedule %s         " % next_playlist.schedule
                s += "\n│      └── Playlist %s         " % next_playlist
                if next_playlist.end_unix > next_playlist.schedule.end_unix:
                    s += "\n│          %s!              " % \
                    (SimpleUtil.red("↑↑↑ Playlist #%s ends after Schedule #%s!" % (next_playlist.playlist_id, next_playlist.schedule.schedule_id)))
                
                entries = self.preprocess_entries(next_playlist.entries, False)
                s += self.build_playlist_string(entries)

        s += "\n└──────────────────────────────────────────────────────────────────────────────────────────────────────\n\n"
        return s


    def build_playlist_string(self, entries):
        """
        Returns a stringified list of entries
        """
        s = ""
        is_out_of_schedule = False

        for entry in entries:
            if entry.queue_state == EntryQueueState.OUT_OF_SCHEDULE and not is_out_of_schedule:
                s += "\n│             %s" % \
                    SimpleUtil.red("↓↓↓ These entries won't be played because they are out of schedule.")
                is_out_of_schedule = True

            s += self.build_entry_string("\n│         └── ", entry, is_out_of_schedule)

        return s


    def build_entry_string(self, prefix, entry, strike):
        """
        Returns an stringified entry.
        """
        s = ""
        if entry.queue_state == EntryQueueState.CUT:
            s = "\n│             %s" % SimpleUtil.red("↓↓↓ This entry is going to be cut.")

        if strike: 
            entry_str = SimpleUtil.strike(entry)
        else:
            entry_str = str(entry)
        
        entry_line = "%sEntry %s | %s" % (prefix, str(entry.entry_num+1), entry_str)
        return s + entry_line



#
#   PRIVATE METHODS
#

    def get_virtual_now(self):
        """
        Liquidsoap is slow in executing commands, therefore it's needed to schedule 
        actions by (n) seconds in advance, as defined in the configuration file by
        the property `lqs_delay_offset`.

        Returns:
            (Integer):  the Unix epoch timestamp including the offset
        """
        time_offset = int(self.config.lqs_delay_offset)
        return SimpleUtil.timestamp() + time_offset



    def get_next_playlists(self):
        """
        Retrieves the playlists to be played after the current one.

        Returns:
            ([Playlist]): The next playlists
        """
        now_unix = self.get_virtual_now()
        next_playlists = []

        for schedule in self.programme:
            if schedule.end_unix > now_unix:
                for playlist in schedule.playlist:
                    if playlist.start_unix > now_unix:
                        next_playlists.append(playlist)

        return next_playlists



    def queue_programme(self):
        """
        Queues the current programme (playlists as per schedule) by creating
        timed commands to Liquidsoap to enable the individual tracks of playlists. 
        """
        current_schedule, current_playlist = self.get_active_playlist()
        playlists = self.get_next_playlists()

        if current_schedule and current_playlist:
            active_entry = self.get_active_entry()
            # Finished entries
            for entry in current_playlist.entries:
                if active_entry == entry:
                    break

            # Entry currently being played
            if active_entry:
                # Open entries for current playlist
                rest_of_playlist = active_entry.get_next_entries(True)
                self.queue_playlist_entries(rest_of_playlist, False, True)

        if playlists:
            for next_playlist in playlists:
               self.queue_playlist_entries(next_playlist.entries, False, True)
        
        self.logger.info(SimpleUtil.green("Finished queuing programme!"))



    def queue_playlist_entries(self, entries, fade_in, fade_out):
        """
        Creates Liquidsoap player commands for all playlist items to be executed at the scheduled time.
        
        Args:
            entries ([PlaylistEntry]):  The playlist entries to be scheduled for playout
            fade_in (Boolean):          Fade-in at the beginning of the set of entries
            fade_out (Boolean):         Fade-out at the end of the set of entries

        Returns:
            (String):                   Formatted string to display playlist entries in log
        """       
        entry_groups = []
        entry_groups.append([])
        previous_entry = None
        index = 0

        # Mark entries which start after the end of their schedule or are cut
        clean_entries = self.preprocess_entries(entries, True)

        # Group all filesystem entries, allowing them to be queued at once
        for entry in clean_entries:
            if previous_entry == None or \
                (previous_entry != None and \
                 previous_entry.type == entry.type and \
                 entry.type == ChannelType.FILESYSTEM):
                
                entry_groups[index].append(entry)
            else:
                index += 1
                entry_groups.append([])
                entry_groups[index].append(entry)
            previous_entry = entry
        self.logger.info("Built %s entry group(s)" % len(entry_groups))
         
        # Schedule function calls
        if len(clean_entries) > 0 and len(entry_groups) > 0:
            for entries in entry_groups:
                if not isinstance(entries, list):
                    raise ValueError("Invalid Entry Group: %s" % str(entries))
                
                self.set_entries_timer(entries, fade_in, fade_out)

                # Check if it's the last item, which needs special handling
                if entries[-1] == clean_entries[-1]:
                    # The end of schedule is the actual end of the track
                    self.queue_end_of_schedule(entries[-1], fade_out)
        else:
            self.logger.warn(SimpleUtil.red("Nothing to schedule..."))



    def set_entries_timer(self, entries, fade_in, fade_out):
        """
        Creates timer for loading and playing one or multiple entries. Existing timers are 
        updated.

        Args:
            entries ([]): List of multiple filesystem entries, or a single entry of other types
        """
        play_timer = self.is_something_planned_at_time(entries[0].start_unix)
        now_unix = self.get_virtual_now()
        diff = entries[0].start_unix - now_unix

        # Play function to be called by timer
        def do_play(entries):
            self.logger.info(SimpleUtil.cyan("=== play('%s') ===" % EngineUtil.get_entries_string(entries)))
            transition_type = TransitionType.INSTANT
            if fade_in:
                transition_type = TransitionType.FADE

            if entries[-1].status != EntryPlayState.READY:
                self.logger.critical(SimpleUtil.red("PLAY: For some reason the entry/entries is not yet ready or could not be loaded (Entries: %s)" % str(entries)))
                # TODO Pro-active fallback handling here

            self.soundsystem.play(entries[0], transition_type)
            self.logger.info(self.get_ascii_programme())


        if play_timer:
            # Check if the Playlist IDs are different
            if play_timer.entries[0].playlist.playlist_id == entries[0].playlist.playlist_id:
                # If not, stop and remove the old timer, create a new one
                self.stop_timer(play_timer)                
            else:
                # If the playlists do not differ => reuse the old timer and do nothing
                self.logger.info("Playlist Entry %s is already scheduled - no new timer created!" % str(entries))
                return
        
        # If nothing is planned at given time, create a new timer
        (entries[0].switchtimer, entries[0].loadtimer) = self.create_timer(diff, do_play, entries, switcher=True)



    def preprocess_entries(self, entries, cut_oos):
        """
        Analyses and marks entries which are going to be cut or excluded.

        Args:
            entries ([PlaylistEntry]):  The playlist entries to be scheduled for playout
            cut_oos (Boolean):  If `True` entries which are 'out of schedule' are not returned

        Returns:
            ([PlaylistEntry]):  The list of processed playlist entries
        """
        clean_entries = []

        for entry in entries:

            if entry.entry_start >= entry.playlist.schedule.schedule_end:
                msg = "Filtered entry (%s) after end-of schedule (%s) ... SKIPPED" % (entry, entry.playlist.schedule)
                self.logger.warn(SimpleUtil.red(msg))
                entry.queue_state = EntryQueueState.OUT_OF_SCHEDULE
            elif entry.end_unix > entry.playlist.schedule.end_unix:
                entry.queue_state = EntryQueueState.CUT
            else:
                entry.queue_state = EntryQueueState.OKAY
            
            if not entry.queue_state == EntryQueueState.OUT_OF_SCHEDULE or not cut_oos:
                clean_entries.append(entry)

        return clean_entries



    def queue_end_of_schedule(self, entry, fade_out):
        """
        Queues a soundsystem action to stop/fade-out the given schedule

        Args:
            entry (PlaylistEntry):  The last entry of the schedule
            fade_out (Boolean):     If the entry should be faded-out
        """
        schedule_end = entry.playlist.schedule.schedule_end
        schedule_end_unix = entry.playlist.schedule.end_unix
        now_unix = self.get_virtual_now()
        fade_out_time = 0

        # Stop function to be called when schedule ends
        def do_stop(entries):
            entry = entries[0]
            self.logger.info(SimpleUtil.cyan("=== stop('%s') ===" % entry))
            transition_type = TransitionType.INSTANT
            if fade_out:
                transition_type = TransitionType.FADE
            self.soundsystem.stop(entry, transition_type)

        if fade_out == True:
            fade_out_time = int(round(float(self.config.get("fade_out_time")))) #FIXME Use float
        
        start_fade_out = schedule_end_unix - now_unix - fade_out_time
        entry.fadeouttimer = self.create_timer(start_fade_out, do_stop, [entry], fadeout=True)

        self.logger.info("Fading out schedule in %s seconds at %s | Last entry: %s" % \
            (str(start_fade_out), str(schedule_end), entry))



    def fetch_new_programme(self):
        """
        Fetch the latest programme from `AuraCalendarService` which stores it to the database.
        After that, the programme is in turn loaded from the database and stored in `self.programme`.
        """

        # Fetch programme from API endpoints
        self.logger.info("Trying to fetch new programe from API endpoints...")
        acs = AuraCalendarService(self.config)
        queue = acs.get_queue()
        acs.start() # start fetching thread
        response = queue.get() # wait for the end
        self.logger.info("... Programme fetch via API done!")

        # Reset last successful fetch state
        lsf = self.last_successful_fetch
        self.last_successful_fetch = None

        if response is None:
            self.logger.warning("Trying to load programme from Engine Database, because AuraCalendarService returned an empty response.")
        elif type(response) is list:
            self.programme = response
            if self.programme is not None and len(self.programme) > 0:
                self.last_successful_fetch = datetime.datetime.now()
                self.logger.info(SimpleUtil.green("Finished fetching current programme from API"))
            if len(self.programme) == 0:
                self.logger.critical("Programme fetched from Steering/Tank has no entries!")
        elif response.startswith("fetching_aborted"):
            self.logger.warning("Trying to load programme from database, because fetching was being aborted from AuraCalendarService! Reason: " + response[16:])
        else:
            self.logger.warning("Trying to load programme from database, because i got an unknown response from AuraCalendarService: " + response)

        # Always load latest programme from the database
        self.last_successful_fetch = lsf
        self.load_programme_from_db()
        self.logger.info(SimpleUtil.green("Finished loading current programme from database (%s schedules)" % str(len(self.programme))))
        for schedule in self.programme:
            self.logger.debug("\tSchedule %s with Playlist %s" % (str(schedule), str(schedule.playlist)))




    def load_programme_from_db(self):
        """
        Loads the programme from Engine's database and enables
        them via `self.enable_entries(..)`. After that, the
        current message queue is printed to the console.
        """
        self.programme = Schedule.select_programme()

        if not self.programme:
            self.logger.critical(SimpleUtil.red("Could not load programme from database. We are in big trouble my friend!"))
            return



    def is_something_planned_at_time(self, given_time):
        """
        Checks for existing timers at the given time.
        """
        for t in self.message_timer:
            if t.entries[0].start_unix == given_time and (t.fadein or t.switcher):
                return t
        return False



    def create_timer(self, diff, func, entries, fadein=False, fadeout=False, switcher=False):
        """
        Creates a new timer for timed execution of mixer commands.

        Args:
            diff (Integer):     The difference in seconds from now, when the call should happen
            func (Function):    The function to call
            entries ([]):       The entries to be scheduled

        """
        if not fadein and not fadeout and not switcher or fadein and fadeout or fadein and switcher or fadeout and switcher:
            raise ValueError("You have to call me with either fadein=true, fadeout=true or switcher=True")
        if not isinstance(entries, list):
            raise ValueError("No list of entries passed!")

        t = CallFunctionTimer(diff=diff, func=func, param=entries, fadein=fadein, fadeout=fadeout, switcher=switcher)
        self.message_timer.append(t)
        t.start()

        if switcher:
            # Pre-roll function to be called by timer
            def do_preroll(entries):
                try:
                    if entries[0].type == ChannelType.FILESYSTEM:
                        self.logger.info(SimpleUtil.cyan("=== preroll_group('%s') ===" % EngineUtil.get_entries_string(entries)))
                        self.soundsystem.preroll_group(entries)
                    else:
                        self.logger.info(SimpleUtil.cyan("=== preroll('%s') ===" % EngineUtil.get_entries_string(entries)))
                        self.soundsystem.preroll(entries[0])
                except LoadSourceException as e:
                    self.logger("Could not load entries %s:" % EngineUtil.get_entries_string(entries), e)
                    # TODO Fallback logic here

            loader_diff = diff - self.config.get("preload_offset")
            loader = CallFunctionTimer(diff=loader_diff, func=do_preroll, param=entries, fadein=fadein, fadeout=fadeout, switcher=False, loader=True)
            self.message_timer.append(loader)
            loader.start()
            return (t, loader)
        else:
            return t



    def stop_timer(self, timer):
        """
        Stops the given timer.

        Args:
            timer (Timer):  The timer to stop.
        """
        timer.cancel()
        count = 1

        if timer.entries[0].loadtimer is not None:
            timer.entries[0].loadtimer.cancel()
            self.message_timer.remove(timer.entries[0].loadtimer)
            count += 1

        # if timer.entries[0].fadeintimer is not None:
        #     timer.entries[0].fadeintimer.cancel()
        #     self.message_timer.remove(timer.entries[0].fadeintimer)
        #     count += 1

        if timer.entries[0].fadeouttimer is not None:
            timer.entries[0].fadeouttimer.cancel()
            self.message_timer.remove(timer.entries[0].fadeouttimer)
            count += 1

        # Remove it from message queue
        self.message_timer.remove(timer)
        self.logger.info("Stopped %s timers for: %s" % (str(count), EngineUtil.get_entries_string(timer.entries)))



    # FIXME Move to adequate module
    def init_database(self):
        """
        Initializes the database.

        Raises:
            sqlalchemy.exc.ProgrammingError:    In case the DB model is invalid
        """
        if self.config.get("recreate_db") is not None:
            AuraDatabaseModel.recreate_db(systemexit=True)

        # Check if tables exists, if not create them
        try:
            Playlist.select_all()
        except sqlalchemy.exc.ProgrammingError as e:
            errcode = e.orig.args[0]

            if errcode == 1146: # Error for no such table
                model = AuraDatabaseModel()
                model.recreate_db()
            else:
                raise



    def terminate(self):
        """
        Called when thread is stopped or a signal to terminate is received.
        """
        self.exit_event.set()
        self.logger.info("Shutting down scheduler ...")


# ------------------------------------------------------------------------------------------ #
# class SetNextFile(threading.Thread):
#     fallbackname = None
#     show = None

#     def __init__(self, fallbackname, show):
#         threading.Thread.__init__(self)
#         self.fallbackname = fallbackname
#         self.show = show

#     def run(self):
#         if self.fallbackname == "show":
#             self.detect_next_file_for(self.show.showfallback)
#         elif self.fallbackname == "timeslow":
#             self.detect_next_file_for(self.show.timeslotfallback)
#         elif self.fallbackname == "station":
#             self.detect_next_file_for(self.show.stationfallback)

#     def detect_next_file_for(self, playlist):
#         return ""
#         #if playlist.startswith("pool"):
#         #    self.find_next_file_in_pool(playlist)

#     #def find_next_file_in_pool(self, pool):
#     #    return ""

# ------------------------------------------------------------------------------------------ #
class CallFunctionTimer(threading.Timer):
    logger = None
    param = None
    entries = None
    diff = None
    fadein = False
    fadeout = False
    switcher = False
    loader = False

    def __init__(self, diff=None, func=None, param=None, fadein=False, fadeout=False, switcher=False, loader=False):

        self.logger = logging.getLogger("AuraEngine")
        self.logger.debug("Executing soundsystem command '%s' in %s seconds..." % (str(func.__name__), str(diff)))
        threading.Timer.__init__(self, diff, func, (param,))

        if not fadein and not fadeout and not switcher and not loader \
            or fadein and fadeout \
            or fadein and switcher \
            or fadeout and switcher:
            
            raise Exception("You have to create me with either fadein=True, fadeout=True or switcher=True")

        self.diff = diff
        self.func = func
        self.param = param
        self.entries = param
        self.fadein = fadein
        self.fadeout = fadeout
        self.switcher = switcher
        self.loader = loader


    def __str__(self):
        """
        String represenation of the timer.
        """
        status = "Timer (Alive: %s)" % self.is_alive()
        time_start = SimpleUtil.fmt_time(SimpleUtil.timestamp() + self.diff)
        status += " starting in " + str(self.diff) + "s (" + time_start + ") "

        if self.fadein:
            return status + " fading in entries '" + EngineUtil.get_entries_string(self.entries)
        elif self.fadeout:
            return status + " fading out schedule '" + str(self.entries[0].playlist.schedule)
        elif self.switcher:
            return status + " switching to entries '" + EngineUtil.get_entries_string(self.entries)
        elif self.loader:
            return status + " pre-rolling entries '" + EngineUtil.get_entries_string(self.entries)
        else:
            return "CORRUPTED CallFunctionTimer around! How can that be?"
